/*
    FEARLUS model0-6-8-2: ParcelLandManagerReport.m
    Copyright (C) 1999-2004, 2007-2008  Macaulay Institute

    This file is part of FEARLUS model0-6-8-2, an agent-based model of land use
    change.

    FEARLUS model0-6-8-2 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-8-2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


The implementation for the ParcelLandManagerReport object,
and for the ParcelsManagersPair object

*/

#import "ParcelLandManagerReport.h"
#import "Parameter.h"
#import "ModelSwarm.h"
#import "FearlusArguments.h"
#import "Environment.h"
#import "LandAllocator.h"
#import "LandManager.h"
#import "MiscFunc.h"
#import "Number.h"



@implementation ParcelLandManagerReport

/*

create:

Creation method. Create the list that will store the number of Land
Managers that own a certain number of Land Parcels

*/

+(id <Report>)create: aZone {
  ParcelLandManagerReport *r;

  r = [super create: aZone];

  r->sort = YES;
 
  return r;
}

/*

setModelSwarm:andParameters:

Set the model swarm and parameters for this report. These will be used
to obtain the data in the report. 

*/

-(void)setModelSwarm: (ModelSwarm *)m andParameters: (Parameter *)p {
  parameter = p;
  model = m;
}

/*

reportForYear:toFile:

Create a report of the number of Land Managers that own a certain number 
of Land Parcels at the end of the year. 

*/

-(void)reportForYear: (unsigned)year toFile: (FILE *)fp {
  id <List> parcelsManagersPairsList;
  id inx, innerInx;
  LandManager *lm;
  ParcelsManagersPair *pmp;
   
  //Create a list of pairs (numberOfLandParcels, numberOfLandManagers)
  //So far the list is empty
  parcelsManagersPairsList = [List create: [self getZone]];

  //Loop through the Land Managers asking them how many Land Parcels 
  //they own.
  for(inx = [[[model getLandAllocator] getLandManagers] begin: [self getZone]],
	[inx next];
      [inx getLoc] == Member;
      [inx next]) {

    int numberOfLandParcels;
    BOOL landManagersParcelsAdded = NO;

    lm = (LandManager *)[inx get];
    numberOfLandParcels = [lm getNumberOfLandParcelsOwned];

    /*Now loop through the list of LandParcels-LandManagers pairs that we have
      so far. If this number of Land Parcels is already on the list, then
      add one to the number of Land Managers that have that number of Land
      Parcels. If not, add the new number of Land Parcels to the list
      (with one Land Manager so far).*/
    
    for(innerInx = [parcelsManagersPairsList begin: [self getZone]],
	[innerInx next];
      [innerInx getLoc] == Member;
      [innerInx next]) {
      
      pmp = (ParcelsManagersPair *)[innerInx get];
      if(numberOfLandParcels == [pmp getNumberOfLandParcels]){
	[pmp addOneToNumberOfLandManagers];
	landManagersParcelsAdded = YES;
	break;
      }
    }
    [innerInx drop];

    //If that number of Land Parcels is new, create a new object and 
    //add it to the parcelsManagersPairsList list.
    if(landManagersParcelsAdded == NO){
      ParcelsManagersPair *newPair;
      newPair = [ParcelsManagersPair create: [self getZone]
				     landParcels: numberOfLandParcels];
      [parcelsManagersPairsList addLast: newPair];
    }
  }
  [inx drop];

  /* Now it's time to print what we've got*/

  //Sort the list if you have to
  if(sort){
    [MiscFunc mergeSort: parcelsManagersPairsList 
	      withSelector: M(getNumberOfLandParcels)];
  }

  for(inx = [parcelsManagersPairsList begin: [self getZone]],
	[inx next];
      [inx getLoc] == Member;
      [inx next]) {
    
    pmp = (ParcelsManagersPair *)[inx get];
    fprintf(fp, "%d\tLand managers own\t%d\tLand Parcels\n", 
	    [pmp getNumberOfLandManagers], [pmp getNumberOfLandParcels]);
  }
  [inx drop];

  
  //We've finished. Now drop everything!
  for(inx = [parcelsManagersPairsList begin: [self getZone]],
	[inx next];
      [inx getLoc] == Member;
      [inx next]) {
    pmp = (ParcelsManagersPair *)[inx get];
    [pmp drop];
  }
  [inx drop];
  [parcelsManagersPairsList drop];  

}

/*

setOption:toValue:

*/

-(BOOL)setOption: (char *)option toValue: (char *)value {
  fprintf(stderr, "Invalid format in report configuration file %s\n"
	  "%s: There are no options for this report: (%s=%s).\n",	
	  [FearlusArguments getReporterCFFile], class_get_class_name([self class]),
	  option, value);
  exit(1);
  return NO;
}

/*

setOptionFlag:toValue:

*/

-(BOOL)setOptionFlag: (char *)option toValue: (BOOL)value {
  if(strcmp(option,"Sorted") == 0){
    sort = value;
  }
  else{
    fprintf(stderr, "Flag %s is not a valid flag for report %s"
	    " in report configuration file %s\n", option, 	
	    class_get_class_name([self class]),
	    [FearlusArguments getReporterCFFile]);
    exit(1);
  }

  return YES;
}

/*

setFixedYears:

Set the fixed years for reporting.

*/

-(void)setFixedYears: (id <List>)y {
  fixedYears = y;
}

/*

setRepeatYears:

Set the repeating years for reporting.

*/

-(void)setRepeatYears: (id <List>)y {
  repeatYears = y;
}

/* There's no need for a drop method -- the Reporter object will own all Report
   objects, and stores them in its own Zone. When the zone is deleted, all the
   Reports' allocated memory will go with it, so long as we use [self getZone]
   here when allocating memory. */

/*

reportingForYear:

Return YES if we're reporting this year, according to the years we've been
handed. NO otherwise.

*/

-(BOOL)reportingForYear: (unsigned)year {
  BOOL reportingYear;
  id inx;
  Number *n;

  /* Decide if we should print this year */
  reportingYear = NO;
  if(([fixedYears getCount] > 0)
     && ([(Number *)[fixedYears atOffset: 0] getUnsigned] == year)) {
    /* This is a fixed year to print. So we'll go ahead, but after we've
       removed the head */
    [[fixedYears removeFirst] drop];
    reportingYear = YES;
  }
  else {
    /* Not a fixed year to print, so check all the repeating years */
    for(inx = [repeatYears begin: scratchZone], [inx next];
	[inx getLoc] == Member;
	[inx next]) {
      n = [inx get];
      if((year % [n getUnsigned]) == 0) {
	/* It's a repeating year to print */
	reportingYear = YES;
	break;
      }
    }
    [inx drop];
  }
  return reportingYear;
}

@end


@implementation ParcelsManagersPair

+(id)create: aZone landParcels: (int)lp
{
  ParcelsManagersPair *obj;

  obj = [super create: aZone];

  obj->numberOfLandParcels = lp;
  obj->numberOfLandManagers = 1;

  return obj;
}

-(int)getNumberOfLandParcels
{
  return numberOfLandParcels;
}

-(int)getNumberOfLandManagers
{
  return numberOfLandManagers;
}

-addOneToNumberOfLandManagers
{
  numberOfLandManagers++;
  return self;
}

@end
