/*
    FEARLUS model0-6-8-2: ParcelWeightedCopyingStrategy.m
    Copyright (C) 1999-2004, 2007-2008  Macaulay Institute

    This file is part of FEARLUS model0-6-8-2, an agent-based model of land use
    change.

    FEARLUS model0-6-8-2 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-8-2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the ParcelWeightedCopyingStrategy object. 

*/

#import "ParcelWeightedCopyingStrategy.h"
#import "SelectUseBucket.h"
#import "LandUse.h"
#import "LandParcel.h"
#import "LandManager.h"
#import "Environment.h"
#import "Parameter.h"
#import "FearlusArguments.h"
#import "BitString.h"

@implementation ParcelWeightedCopyingStrategy

/*

create:withManager:andParameters:

Create the strategy, setting the manager and parameters.

*/

+(id <ImitativeStrategy, NonHistoricalStrategy>)create: aZone
	   withManager: (LandManager *)lmgr
	 andParameters: (Parameter *)p {
  ParcelWeightedCopyingStrategy *obj;

  obj = [super create: aZone];
  obj->lm = lmgr;
  obj->parameter = p;
  return obj;
}

/*

decideLandUseForParcel:

Decide the land use for the land parcel, using a simple copying strategy.
Create a bucket, which is filled with m * neighbourhood weighting of the
land manager for each land use used by a parcel owned by a neighbour of the
land manager, and m for each land use used by a parcel owned by the land
manager, where m is the match between the bit string of the land parcel with
that of the potentially copied land parcel.

Neighbourhood noise is applied to the bit string of any neighbouring land
parcel before computing the match.

*/

-(LandUse *)decideLandUseForParcel: (LandParcel *)lp {
  SelectUseBucket *bucket;
  id parcelIndex, tempZone;
  LandParcel *myParcel, *neighbouringParcel;
  LandManager *neighbour;
  LandUse *lu;
  BitString *bs;
  double match;
  int i;

  tempZone = [Zone create: scratchZone];
  bucket = [SelectUseBucket create: tempZone
			    withParameters: parameter
			    andLandUses: [[lp getEnvironment] getLandUses]];
  // Loop through land parcels owned by the land manager
  if([FearlusArguments debugCoreMethods]) {
    printf("S: %s: Using parcel weighted copying strategy of LM = land manager"
	   " %u to determine land use\n", sel_get_name(_cmd), [lm getPIN]);
    printf("S: %s: Adding scores for match of LP = land parcel %u at (%d, "
	   "%d) with other land parcels owned by LM\n", sel_get_name(_cmd),
	   [lp getPIN], [lp getX], [lp getY]);
  }
  for(parcelIndex = [[lm getLandParcels] begin: tempZone], [parcelIndex next];
      [parcelIndex getLoc] == Member;
      [parcelIndex next]) {
    myParcel = [parcelIndex get];
    match = (double)[ [lp getAttributes]
		      getMatchWith: [myParcel getAttributes] ];
    [bucket addScore: match toBucketForLandUse: [myParcel getLandUse]];
    [myParcel incNImitations];
    if([FearlusArguments debugCoreMethods]) {
      printf("M: %s: Match of LP with land parcel %u at (%d, %d) with land use"
	     " %u = %g\n", sel_get_name(_cmd), [myParcel getPIN],
	     [myParcel getX], [myParcel getY], [[myParcel getLandUse] getPIN],
	     match);
    }
  }
  [parcelIndex drop];
  // Loop through the neighbours, and through their land parcels
  bs = [BitString create: tempZone
		  setBitCount: [parameter parcelBitStringSize]];
  for(i = 0; (neighbour = [lm getNextNeighbour: i]); i |= 1) {
    if([FearlusArguments debugCoreMethods]) {
      printf("S: %s: Adding scores for match of LP with land parcels belonging"
	     " to land manager %u\n", sel_get_name(_cmd), [neighbour getPIN]);
    }
    for(parcelIndex = [[neighbour getLandParcels] begin: tempZone],
	  [parcelIndex next];
	[parcelIndex getLoc] == Member;
	[parcelIndex next]) {
      neighbouringParcel = [parcelIndex get];
      [bs setToCopyOf: [neighbouringParcel getAttributes]];
      [bs mutateWithProbability: [lm getNeighbourNoise]];
      match = (double)[ [lp getAttributes] getMatchWith: bs ];
      [bucket addScore: [lm getNeighbourWeight] * match
	      toBucketForLandUse: [neighbouringParcel getLandUse]];
      [neighbouringParcel incNImitations];
      if([FearlusArguments debugCoreMethods]) {
	printf("S: %s: Bit string seen by LM for neighbour %u's land parcel %u"
	       " at (%d, %d) = ", sel_get_name(_cmd), [neighbour getPIN],
	       [neighbouringParcel getPIN], [neighbouringParcel getX],
	       [neighbouringParcel getY]);
	[bs printToFile: stdout];
	printf(" (%d bits mutated)\n", (int)[bs getBitCount] -
	       (int)[[neighbouringParcel getAttributes] getMatchWith: bs]);
	printf("S: %s: Match of LP with land parcel %u at (%d, %d) with land "
	       "use %u = %g. Score of land use incremented by %g\n",
	       sel_get_name(_cmd), [neighbouringParcel getPIN],
	       [neighbouringParcel getX], [neighbouringParcel getY],
	       [[neighbouringParcel getLandUse] getPIN], match,
	       match * [lm getNeighbourWeight]);
      }
    }
    [parcelIndex drop];
  }
  lu = [bucket getChoice];
  [bucket drop];
  [bs drop];
  [tempZone drop];
  return lu;
}

@end
